@@ -7,7 +7,7 @@
 #  the U.S. Government retains certain rights in this software.
 #  _________________________________________________________________________
 
-__all__ = ['subprocess', 'SubprocessMngr', 'run_command', 'timer', 'signal_handler', 'run']
+__all__ = ['subprocess', 'SubprocessMngr', 'run_command', 'timer', 'signal_handler', 'run', 'PIPE', 'STDOUT']
 
 from pyutilib.subprocess import GlobalData
 import time
@@ -17,10 +17,22 @@ import sys
 import tempfile
 import subprocess
 import copy
+from six import itervalues
 from threading import Thread
 if subprocess.mswindows and sys.version_info[0:2] < (2,5):
     import ctypes
 
+_peek_available = True
+try:
+    if subprocess.mswindows:
+        from msvcrt import get_osfhandle
+        from win32pipe import PeekNamedPipe
+        from win32file import ReadFile
+    else:
+        from select import select
+except:
+    _peek_available = False
+
 import pyutilib.services
 from pyutilib.common import *
 from pyutilib.misc import quote_split
@@ -30,6 +42,9 @@ try:
 except:
     basestring = unicode = str
 
+PIPE = subprocess.PIPE
+STDOUT = subprocess.STDOUT
+
 if sys.version_info[0:2] >= (3,0):
     def bytes_cast(x):
         if not x is None:
@@ -130,27 +145,165 @@ def signal_handler(signum, frame, verbose=False):
 #
 # A function used to read in data from a shell command, and push it into a pipe.
 #
-def _read_data(stream, ostr1, ostr2):
+def _stream_reader(args):
+    unbuffer = args[0]
+    stream = args[1]
+    write = tuple( x for x in args[2:] if x is not None )
+
+    encoding = sys.__stdout__.encoding
+    if encoding is None:
+        encoding = 'utf-8'
+
+    buf = ""
+    data = None
+
     while True:
-        data = os.read(stream, 80)
-        if sys.__stdout__.encoding is None:
-            data = data.decode('utf-8')
-        else:
-            data = data.decode(sys.__stdout__.encoding)
-        if data:
-            ostr1.write(data)
-            ostr1.flush()
-            ostr2.write(data)
-            ostr2.flush()
-        else:
+        new_data = os.read(stream, 1)
+        if not new_data:
             break
+        if data:
+            data += new_data
+        else:
+            data = new_data
+        char = data.decode(encoding)
+        if char.encode(encoding) != data:
+            continue
+        data = ""
+        if unbuffer == 1:
+            for s in write:
+                s.write(char)
+        buf += char
+        if char[-1] != "\n":
+            continue
+        if unbuffer:
+            if unbuffer != 1:
+                for s in write:
+                    s.write(buf)             
+            for s in write:
+                s.flush()
+        else:
+            for s in write:
+                s.write(buf)
+        buf = ""
+    if buf:
+        for s in write:
+            s.write(buf)
+    if data:
+        for s in write:
+            s.write(data.decode(encoding))
+    for s in write:
+        s.flush()
+        
+#
+# A function used to read in data from two independent streams and push
+# each to 1+ output pipes.  Managing this in a single thread allows our
+# code output to be more predictible (and better formatted) when the
+# output streams happen to point to the same place (e.g., file or
+# terminal).  This function requires the ability to "peek" or "select"
+# the next stream that is ready to be read from.
+#
+# For platforms that do not support select / peek, see the
+# _pseudo_merged_reader.
+#
+def _merged_reader(*args):
+    class StreamData(object):
+        __slots__ = ( 'read','write','unbuffer','buf','data' )
+        def __init__(self, *args):
+            if subprocess.mswindows:
+                self.read = get_osfhandle( args[1] )
+            else:
+                self.read = args[1]
+            self.unbuffer = args[0]
+            self.write = tuple( x for x in args[2:] if x is not None )
+            self.buf = ""
+            self.data = None
+        
+    encoding = sys.__stdout__.encoding
+    if encoding is None:
+        encoding = 'utf-8'
+
+    streams = {}
+    for s in args:
+        tmp = StreamData(*s)
+        streams[tmp.read] = tmp
+
+    handles = sorted(streams.keys(), key=lambda x: -1*streams[x].unbuffer)
+    noop = []
+
+    while handles:
+        if subprocess.mswindows:
+            for h in handles:
+                numAvail = PeekNamedPipe(h, 0)[1]
+                if numAvail == 0:
+                    if h.closed:
+                        handles.remove(h)
+                    continue
+                new_data = ReadFile(h, 1, None)
+        else:
+            h = select(handles,noop,noop)[0]
+            if not h:
+                break
+            h = h[0]
+            new_data = os.read(h, 1)
+            if not new_data:
+                handles.remove(h)
+                continue
+        s = streams[h]
+        if s.data is None:
+            s.data = new_data
+        else:
+            s.data += new_data
+        char = s.data.decode(encoding)
+        if char.encode(encoding) != s.data:
+            continue
+        s.data = None
+        if s.unbuffer:
+            for x in s.write:
+                x.write(char)
+        s.buf += char
+        if char[-1] != "\n":
+            continue
+        if s.unbuffer:
+            for x in s.write:
+                x.flush()
+        else:
+            for x in s.write:
+                x.write(s.buf)
+        s.buf = ""
+    for s in itervalues(streams):
+        if s.buf:
+            for x in s.write:
+                x.write(s.buf)
+        if s.data:
+            tmp = s.data.decode(encoding)
+            for x in s.write:
+                x.write(tmp)
+        for x in s.write:
+            x.flush()
+        
+#
+# A mock-up of the _merged_reader for platforms (or installations) that
+# lack working select/peek implementations.  Note that this function
+# does not guarantee determinism (and on many platforms is exceedingly
+# nondeterministic).  However, it does change the flushing rules to
+# better maintain output line integrity (at the cost of performance).
+#
+def _pseudo_merged_reader(*args):
+    _threads = []
+    for arg in args:
+        _threads.append(Thread( target=_stream_reader, args=(( 2, )+arg[1:],) ))
+        _threads[-1].daemon = True
+        _threads[-1].start()
+    for th in _threads:
+        th.join()
+
 
 #
 # Execute the command as a subprocess that we can send signals to.
 # After this is finished, we can get the output from this command from
 # the process.stdout file descriptor.
 #
-def run_command(cmd, outfile=None, cwd=None, ostream=None, stdin=None, stdout=None, stderr=None, valgrind=False, valgrind_log=None, valgrind_options=None, memmon=False, env=None, define_signal_handlers=True, debug=False, verbose=True, timelimit=None, tee=None, ignore_output=False, shell=False, thread_reader=_read_data):
+def run_command(cmd, outfile=None, cwd=None, ostream=None, stdin=None, stdout=None, stderr=None, valgrind=False, valgrind_log=None, valgrind_options=None, memmon=False, env=None, define_signal_handlers=True, debug=False, verbose=True, timelimit=None, tee=None, ignore_output=False, shell=False, thread_reader=None):
     #
     # Move to the specified working directory
     #
@@ -202,28 +355,39 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdin=None, stdout=No
     #
     tmpfile=None
     if ostream is not None:
-        stdout_arg=ostream
-        stderr_arg=ostream
-        if outfile is None:
-            output = "Output printed to specified ostream"
-        else:
-            output = "Output printed to specified ostream (%s)" % outfile
+        stdout_arg = stderr_arg = ostream
+        if outfile is not None or stdout is not None or stderr is not None:
+            raise ValueError(
+                "subprocess.run_command(): ostream, outfile, and "
+                "{stdout, stderr} options are mutually exclusive" )
+        output = "Output printed to specified ostream"
     elif outfile is not None:
-        stdout_arg=open(outfile,"w")
-        stderr_arg=stdout_arg
+        stdout_arg = stderr_arg = open(outfile,"w")
+        if stdout is not None or stderr is not None:
+            raise ValueError(
+                "subprocess.run_command(): outfile and "
+                "{stdout, stderr} options are mutually exclusive" )
         output = "Output printed to file '%s'" % outfile
     elif not (stdout is None and stderr is None):
         stdout_arg=stdout
         stderr_arg=stderr
         output = "Output printed to specified stdout and stderr streams"
     else:
-        # Create a temporary file.  The mode is w+, which means that we can read and write.
-        # NOTE: the default mode is w+b, but writing to the binary mode seems to cause problems
-        # in the _read_data function on Python 3.x.
-        tmpfile = tempfile.TemporaryFile(mode='w+')
-        stdout_arg=tmpfile
-        stderr_arg=tmpfile
+        # Create a temporary file.  The mode is w+, which means that we
+        #   can read and write.
+        # NOTE: the default mode is w+b, but writing to the binary mode
+        #   seems to cause problems in the _stream_reader function on Python
+        #   3.x.
+        stdout_arg = stderr_arg = tmpfile = tempfile.TemporaryFile(mode='w+')
         output=""
+
+    if stdout_arg is stderr_arg:
+        try:
+            if not tee or ( not tee[0] and not tee[1] ):
+                stderr_arg = STDOUT
+        except:
+            pass
+
     #
     # Setup the default environment
     #
@@ -247,12 +411,22 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdin=None, stdout=No
     if debug:
         print("Executing command %s" % (_cmd, ))
     try:
+        try:
+            simpleCase = not tee
+            if stdout_arg is not None:
+                stdout_arg.fileno()
+            if stderr_arg is not None:
+                stderr_arg.fileno()
+        except:
+            simpleCase = False
+
         GlobalData.signal_handler_busy=False
-        if (tee is None) or (tee is False):
+        if simpleCase:
             #
             # Redirect IO to the stdout_arg/stderr_arg files
             #
-            process = SubprocessMngr(_cmd, stdin=stdin, stdout=stdout_arg, stderr=stderr_arg, env=env, shell=shell)
+            process = SubprocessMngr( _cmd, stdin=stdin, stdout=stdout_arg, 
+                                      stderr=stderr_arg, env=env, shell=shell )
             GlobalData.current_process = process.process
             rc = process.wait(timelimit)
             GlobalData.current_process = None
@@ -262,32 +436,96 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdin=None, stdout=No
             # send this to both the stdout/stdarg value, as well
             # as doing a normal 'print'
             #
-            rpipe1, wpipe1 = os.pipe()
+            out_fd = []
+            out_th = []
+            for fid in (0,1):
+                if fid == 0:
+                    s, raw = stdout_arg, sys.stdout
+                else: 
+                    s, raw = stderr_arg, sys.stderr
+                try:
+                    tee_fid = tee[fid]
+                except:
+                    tee_fid = tee
+                if s is None or s is STDOUT:
+                    out_fd.append(s)
+                elif not tee_fid:
+                    # This catches using StringIO as an output buffer:
+                    # Python's subprocess requires the stream objects to
+                    # have a "fileno()" attribute, which StringIO does
+                    # not have.  We will mock things up by putting a
+                    # pipe in between the subprocess and the StringIO
+                    # buffer.  <sigh>
+                    #
+                    #if hasattr(s, 'fileno'):
+                    #
+                    # Update: in Python 3, StringIO declares a fileno()
+                    # method, but that method throws an exception.  So,
+                    # we can't just check for the attribute: we *must*
+                    # call the method and see if we get an exception.
+                    try:
+                        s.fileno() 
+                        out_fd.append(s)
+                    except:
+                        r, w = os.pipe()
+                        out_fd.append(w)
+                        out_th.append(((fid,r,s), r, w))
+                        #th = Thread(target=thread_reader, args=(r,None,s,fid))
+                        #out_th.append((th, r, w))                        
+                else:
+                    r, w = os.pipe()
+                    out_fd.append(w)
+                    out_th.append(((fid,r,raw,s), r, w))
+                    #th = Thread( target=thread_reader, args=(r,raw,s,fid) )
+                    #out_th.append((th, r, w))
             #
-            process = SubprocessMngr(_cmd, stdin=stdin, stdout=wpipe1, stderr=wpipe1, env=env, shell=shell)
+            process = SubprocessMngr(_cmd, stdin=stdin, stdout=out_fd[0], 
+                                     stderr=out_fd[1], env=env, shell=shell)
             GlobalData.current_process = process.process
+            GlobalData.signal_handler_busy=False
             #
             # Create a thread to read in stdout and stderr data
             #
-            GlobalData.signal_handler_busy=False
-            outt = Thread(target=_read_data, args=(rpipe1, sys.stdout, stdout_arg))
-            outt.daemon = True
-            outt.start()
+            if out_th:
+                if thread_reader is not None:
+                    reader = thread_reader
+                elif len(out_th) == 1:
+                    reader = _stream_reader
+                elif _peek_available:
+                    reader = _merged_reader
+                else:
+                    reader = _pseudo_merged_reader
+                th = Thread( target=reader, args=[x[0] for x in out_th] )
+                th.daemon = True
+                th.start()
+            #for th in out_th:
+            #    th[0].daemon = True
+            #    th[0].start()
             #
             # Wait for process to finish
             #
             rc = process.wait(timelimit)
             GlobalData.current_process = None
+            out_fd = None
             #
             # 'Closing' the PIPE to send EOF to the reader.
             #
-            os.close(wpipe1)
-            #
-            # Wait for readers to finish up with the data in the pipe.
-            #
-            outt.join()
-            del outt
-            os.close(rpipe1)
+            if out_th:
+                for p in out_th:
+                    os.close(p[2])
+                th.join()
+                for p in out_th:
+                    os.close(p[1])
+                del th
+            #for th in reversed(out_th):
+            #    os.close(th[2])
+            #    #
+            #    # Wait for readers to finish up with the data in the pipe.
+            #    #
+            #    th[0].join()
+            #    os.close(th[1])
+            #    thread = th[0]
+            #    del thread
 
     except WindowsError:
         err = sys.exc_info()[1]
@@ -299,9 +537,9 @@ def run_command(cmd, outfile=None, cwd=None, ostream=None, stdin=None, stdout=No
         # Ignore IOErrors, which are caused by interupts
         #
         pass
-    if ostream is None and outfile is not None:
+    if outfile is not None:
         stdout_arg.close()
-    elif not tmpfile is None and not ignore_output:
+    elif tmpfile is not None and not ignore_output:
         tmpfile.seek(0)
         output = "".join(tmpfile.readlines())
         tmpfile.close()
@@ -339,8 +577,8 @@ class SubprocessMngr(object):
         #
         # By default, stderr is mapped to stdout
         #
-        if stderr is None:
-            stderr=subprocess.STDOUT
+        #if stderr is None:
+        #    stderr=subprocess.STDOUT
 
         self.stdin = stdin
         if stdin is None:
